/* 
 *  XMMS Crossfade Plugin
 *  Copyright (C) 2000  Peter Eisenlohr <p.eisenlohr@gmx.net>
 *
 *  based on the original OSS Output Plugin
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 *  USA.
 */

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#elif defined(HAVE_MACHINE_SOUNDCARD_H)
#include <machine/soundcard.h>
#endif

#include "xmms/plugin.h"
#include "xmms/xmmsctrl.h"
#include "xmms/dirbrowser.h"
#include "xmms/configfile.h"
#include "xmms/util.h"

#include "crossfade.h"
#include "configure.h"
#include "monitor.h"

#include "interface.h"
#include "support.h"


#define SET_SENSITIVE(name, sensitive)		\
if((set_wgt = lookup_widget(config_win, name)))	\
  gtk_widget_set_sensitive(set_wgt, sensitive)

#define SET_TOGGLE(name, active)					\
if((set_wgt = lookup_widget(config_win, name)))				\
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(set_wgt), active)

#define SET_SPIN(name, value)					\
if((set_wgt = lookup_widget(config_win, name)))			\
  gtk_spin_button_set_value(GTK_SPIN_BUTTON(set_wgt), value);

#define SET_PAGE(name, index)				\
if((set_wgt = lookup_widget(config_win, name)))		\
  gtk_notebook_set_page(GTK_NOTEBOOK(set_wgt), index);

#define SET_HISTORY(name, index)				\
if((set_wgt = lookup_widget(config_win, name)))			\
  gtk_option_menu_set_history(GTK_OPTION_MENU(set_wgt), index);


#define GET_SENSITIVE(name)			\
((get_wgt = lookup_widget(config_win, name))	\
  && GTK_WIDGET_SENSITIVE(get_wgt))		\

#define GET_TOGGLE(name)					\
((get_wgt = lookup_widget(config_win, name))			\
  && gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(get_wgt)))

#define GET_SPIN(name)							\
((get_wgt = lookup_widget(config_win, name))				\
  ? gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(get_wgt)) : 0)


static GtkWidget *config_win = NULL;
static GtkWidget *about_win = NULL;
static GtkWidget *set_wgt;
static GtkWidget *get_wgt;

static gint output_method;

static gint oss_audio_device;
static gint oss_mixer_device;

static gchar          *op_name = NULL;
static gchar          *op_config_string = NULL;
static gint            op_index;
static plugin_config_t op_config;

static gboolean        xf_checking_dependencies = FALSE;
static fade_config_t   xf_config[MAX_FADE_CONFIGS];
static gint            xf_index;

static gboolean unlocked_gap_trail_enable;
static gint     unlocked_gap_trail_len_ms;
static gint     unlocked_gap_trail_level;


static void update_plugin_options(gchar **config_string, gchar *name,
				  plugin_config_t *pc, gboolean save);

/*****************************************************************************/

static void
read_fade_config(ConfigFile *cfgfile, gchar *key, fade_config_t *fc)
{
  gchar *s = NULL;
  gint n;
  
  if(!cfgfile || !key || !fc) return;

  xmms_cfg_read_string(cfgfile, "Crossfade", key, &s);
  if(!s) return;

  n = sscanf(s,
	     "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
	     &fc->type,
	     &fc->pause_len_ms,
	     &fc->simple_len_ms,
	     &fc->out_enable,
	     &fc->out_len_ms,
	     &fc->out_volume,
	     &fc->ofs_type,
	     &fc->ofs_type_wanted,
	     &fc->ofs_custom_ms,
	     &fc->in_locked,
	     &fc->in_enable,
	     &fc->in_len_ms,
	     &fc->in_volume);

  g_free(s);
}

static void
write_fade_config(ConfigFile *cfgfile, gchar *key, fade_config_t *fc)
{
  gchar *s;
  
  if(!cfgfile || !key || !fc) return;

  s = g_strdup_printf("%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d",
		      fc->type,
		      fc->pause_len_ms,
		      fc->simple_len_ms,
		      fc->out_enable,
		      fc->out_len_ms,
		      fc->out_volume,
		      fc->ofs_type,
		      fc->ofs_type_wanted,
		      fc->ofs_custom_ms,
		      fc->in_locked,
		      fc->in_enable,
		      fc->in_len_ms,
		      fc->in_volume);
  if(!s) return;

  xmms_cfg_write_string(cfgfile, "Crossfade", key, s);
  g_free(s);
}

void
xfade_load_config()
{
  gchar      *filename;
  ConfigFile *cfgfile;

  if(!(filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL))) {
    DEBUG(("[crossfade] save_config: g_strconcat() failed!\n"));
    return;
  }

  if((cfgfile = xmms_cfg_open_file(filename))) {
    /* config items used in v0.1 */
    xmms_cfg_read_int    (cfgfile, "Crossfade", "output_method",        &config->output_method);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "audio_device",         &config->oss_audio_device);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "use_alt_audio_device", &config->oss_use_alt_audio_device);
    xmms_cfg_read_string (cfgfile, "Crossfade", "alt_audio_device",     &config->oss_alt_audio_device);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "mixer_device",         &config->oss_mixer_device);
    xmms_cfg_read_string (cfgfile, "Crossfade", "output_plugin",        &config->op_name);
    xmms_cfg_read_string (cfgfile, "Crossfade", "op_config_string",     &config->op_config_string);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "buffer_size",          &config->mix_size_ms);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "sync_size",            &config->sync_size_ms);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "preload_size",         &config->preload_size_ms);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "songchange_timeout",   &config->songchange_timeout);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "enable_mixer",         &config->enable_mixer);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "mixer_reverse",        &config->mixer_reverse);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "enable_debug",         &config->enable_debug);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "enable_monitor",       &config->enable_monitor);

    /* config items introduced by v0.2 */
    xmms_cfg_read_int    (cfgfile, "Crossfade", "oss_buffer_size",      &config->oss_buffer_size_ms);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "oss_preload_size",     &config->oss_preload_size_ms);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "oss_mixer_use_master", &config->oss_mixer_use_master);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "gap_lead_enable",      &config->gap_lead_enable);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "gap_lead_len_ms",      &config->gap_lead_len_ms);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "gap_lead_level",       &config->gap_lead_level);
    xmms_cfg_read_boolean(cfgfile, "Crossfade", "gap_trail_enable",     &config->gap_trail_enable);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "gap_trail_len_ms",     &config->gap_trail_len_ms);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "gap_trail_level",      &config->gap_trail_level);
    xmms_cfg_read_int    (cfgfile, "Crossfade", "gap_trail_locked",     &config->gap_trail_locked);

    read_fade_config(cfgfile, "fc_xfade",  &config->fc[FADE_CONFIG_XFADE]);
    read_fade_config(cfgfile, "fc_manual", &config->fc[FADE_CONFIG_MANUAL]);

    xmms_cfg_free(cfgfile);
    DEBUG(("[crossfade] load_config: configuration loaded (\"%s\")\n", filename));
  }
  else
    DEBUG(("[crossfade] load_config: error loading config, using defaults\n"));

  g_free(filename);
}

void xfade_save_config()
{
  gchar      *filename;
  ConfigFile *cfgfile;

  if(!(filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL))) {
    DEBUG(("[crossfade] save_config: g_strconcat() failed!\n"));
    return;
  }

  cfgfile = xmms_cfg_open_file(filename);
  if(!cfgfile) cfgfile = xmms_cfg_new();
  if(cfgfile) {
    /* config items obsoleted by v0.2 */
    xmms_cfg_remove_key(cfgfile, "Crossfade", "underrun_pct");
    xmms_cfg_remove_key(cfgfile, "Crossfade", "enable_crossfade");
    xmms_cfg_remove_key(cfgfile, "Crossfade", "enable_gapkiller");
    xmms_cfg_remove_key(cfgfile, "Crossfade", "mixer_use_master");
    xmms_cfg_remove_key(cfgfile, "Crossfade", "late_effect");

    /* config items used in v0.1 */
    xmms_cfg_write_int    (cfgfile, "Crossfade", "output_method",        config->output_method);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "audio_device",         config->oss_audio_device);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "use_alt_audio_device", config->oss_use_alt_audio_device);
    xmms_cfg_write_string (cfgfile, "Crossfade", "alt_audio_device",     config->oss_alt_audio_device ? config->oss_alt_audio_device : DEFAULT_OSS_ALT_AUDIO_DEVICE);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "mixer_device",         config->oss_mixer_device);
    xmms_cfg_write_string (cfgfile, "Crossfade", "output_plugin",        config->op_name ? config->op_name : DEFAULT_OP_NAME);
    xmms_cfg_write_string (cfgfile, "Crossfade", "op_config_string",     config->op_config_string ? config->op_config_string : DEFAULT_OP_CONFIG_STRING);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "buffer_size",          config->mix_size_ms);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "sync_size",            config->sync_size_ms);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "preload_size",         config->preload_size_ms);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "songchange_timeout",   config->songchange_timeout);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "enable_mixer",         config->enable_mixer);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "mixer_reverse",        config->mixer_reverse);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "enable_debug",         config->enable_debug);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "enable_monitor",       config->enable_monitor);

    /* config items introduced by v0.2 */
    xmms_cfg_write_int    (cfgfile, "Crossfade", "oss_buffer_size",      config->oss_buffer_size_ms);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "oss_preload_size",     config->oss_preload_size_ms);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "oss_mixer_use_master", config->oss_mixer_use_master);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "gap_lead_enable",      config->gap_lead_enable);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "gap_lead_length",      config->gap_lead_len_ms);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "gap_lead_level",       config->gap_lead_level);
    xmms_cfg_write_boolean(cfgfile, "Crossfade", "gap_trail_enable",     config->gap_trail_enable);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "gap_trail_len_ms",     config->gap_trail_len_ms);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "gap_trail_level",      config->gap_trail_level);
    xmms_cfg_write_int    (cfgfile, "Crossfade", "gap_trail_locked",     config->gap_trail_locked);

    write_fade_config(cfgfile, "fc_xfade",  &config->fc[FADE_CONFIG_XFADE]);
    write_fade_config(cfgfile, "fc_manual", &config->fc[FADE_CONFIG_MANUAL]);

    xmms_cfg_write_file   (cfgfile, filename);
    xmms_cfg_free         (cfgfile);
    DEBUG(("[crossfade] save_config: configuration saved (\"%s\")\n", filename));
  }
  else
    DEBUG(("[crossfade] save_config: error saving configuration!\n"));

  g_free(filename);
}

void xfade_check_config()
{
  /* load plugin options */
  update_plugin_options(&config->op_config_string, config->op_name,
			&config->op_config, FALSE);
}

void xfade_free_config()
{
  if(op_name) { g_free(op_name); op_name = NULL; }
  if(op_config_string) { g_free(op_config_string); op_config_string = NULL; }
}

/*** helpers *****************************************************************/

gboolean
xfade_cfg_use_xfade (fade_config_t *fc)
{
  if(!fc) return FALSE;
  switch(fc->type) {
  case FADE_TYPE_NONE:
  case FADE_TYPE_PAUSE:
  case FADE_TYPE_SIMPLE_XF:
  case FADE_TYPE_ADVANCED_XF:
    return TRUE;
  }
  return FALSE;
}

gint
xfade_cfg_fadeout_len(fade_config_t *fc)
{
  if(!fc) return 0;
  switch(fc->type) {
  case FADE_TYPE_SIMPLE_XF:
    return fc->simple_len_ms;
  case FADE_TYPE_ADVANCED_XF:
    return fc->out_enable ? fc->out_len_ms : 0;
  }
  return 0;
}
gint
xfade_cfg_fadeout_volume(fade_config_t *fc)
{
  gint volume;

  if(!fc || fc->type != FADE_TYPE_ADVANCED_XF) return 0;
  volume = fc->out_volume;
  if(volume <   0) volume =   0;
  if(volume > 100) volume = 100;
  return volume;
}

gint
xfade_cfg_offset(fade_config_t *fc)
{
  if(!fc) return 0;
  switch(fc->type) {
  case FADE_TYPE_PAUSE:
    return fc->pause_len_ms;
  case FADE_TYPE_SIMPLE_XF:
    return -fc->simple_len_ms;
  case FADE_TYPE_ADVANCED_XF:
    switch(fc->ofs_type) {
    case FC_OFFSET_LOCK_OUT:
      return -fc->out_len_ms;
    case FC_OFFSET_LOCK_IN:
      return -fc->in_len_ms;
    case FC_OFFSET_CUSTOM:
      return fc->ofs_custom_ms;
    }
  }
  return 0;
}

gint
xfade_cfg_fadein_len(fade_config_t *fc)
{
  if(!fc) return 0;
  switch(fc->type) {
  case FADE_TYPE_SIMPLE_XF:
    return fc->simple_len_ms;
  case FADE_TYPE_ADVANCED_XF:
    return
      fc->in_locked
      ? (fc->out_enable ? fc->out_len_ms : 0)
      : (fc->in_enable  ? fc->in_len_ms  : 0);
  }
  return 0;
}

gint
xfade_cfg_fadein_volume(fade_config_t *fc)
{
  gint volume;

  if(!fc || fc->type != FADE_TYPE_ADVANCED_XF) return 0;
  volume = fc->in_locked ? fc->out_volume : fc->in_volume;
  if(volume <   0) volume =   0;
  if(volume > 100) volume = 100;
  return volume;
}

/*** internal helpers ********************************************************/

static void
add_menu_item(GtkWidget *menu, gchar *title, GtkSignalFunc func, gint index)
{
  GtkWidget *item;
  if(!menu || !title || !func) return;
  item = gtk_menu_item_new_with_label(title);
  gtk_signal_connect(GTK_OBJECT(item), "activate", func, (gpointer)index);
  gtk_widget_show(item);
  gtk_menu_append(GTK_MENU(menu), item);
}

/*** output method ***********************************************************/

/*-- callbacks --------------------------------------------------------------*/

void on_output_oss_radio_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  SET_PAGE("output_notebook", 0);
  output_method = OUTPUT_METHOD_BUILTIN_OSS;
}

void on_output_plugin_radio_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  SET_PAGE("output_notebook", 1);
  output_method = OUTPUT_METHOD_PLUGIN;
}

void on_output_none_radio_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  SET_PAGE("output_notebook", 2);
  output_method = OUTPUT_METHOD_BUILTIN_NULL;
}

/*** oss output **************************************************************/

static void
scan_devices(gchar *type, GtkWidget *option_menu, GtkSignalFunc signal_f)
{
  gchar buffer[256];
  FILE *file;

  GtkWidget *menu;
  GtkWidget *item;

  gboolean  found = FALSE;
  gint      index = 0;

  menu = gtk_menu_new();

  /* look for devices in /dev/sndstat or /proc/asound/sndstat (OSS style) */
  if((file = fopen("/dev/sndstat", "r")) ||
     (file = fopen("/proc/asound/sndstat", "r"))) {
    while(fgets(buffer, sizeof(buffer), file)) {
      gint i = strlen(buffer)-1;
      while((i >= 0) && isspace(buffer[i])) buffer[i--] = 0;
      if(found) {
	if(!buffer[0] || !isdigit(buffer[0])) break;
	if(index == 0) {
	  gchar *label, *p = strchr(buffer, ':');
	  if(p) do p++; while(*p == ' ');
	  else p = buffer;
	  label = g_strdup_printf("Default (%s)", p);
	  item = gtk_menu_item_new_with_label(label);
	  g_free(label);
	}
	else item = gtk_menu_item_new_with_label(buffer);
	
	gtk_signal_connect(GTK_OBJECT(item), "activate", signal_f,
			   (gpointer)index);
	gtk_widget_show(item);
	gtk_menu_append(GTK_MENU(menu), item);
	index++;
      }
      else if(!strcmp(buffer, type)) found = TRUE;
    }
    fclose(file);
    
    if(!found)
      DEBUG(("[crossfade] scan_devices: section \"%s\" not found!\n", type));
  }
  else {
    DEBUG(("[crossfade] scan_devices: error opening "
	   "/dev/sndstat || /proc/asound/sndstat:\n"));
#ifdef SOUND_MIXER_INFO
    /* from xmms-3dse7 by Frank Cornelis */
    DEBUG(("[crossfade] scan_devices: using alternate method...\n"));
    for(;;) {
      gchar      dev_name[32];
      int        fd;
      mixer_info info;
      gchar     *label;
      
      if(index != 0)
	sprintf(dev_name, "/dev/mixer%d", index);
      else
	strcpy(dev_name, "/dev/mixer");
      
      if((fd = open(dev_name, O_RDONLY)) != -1) {
	if(ioctl(fd, SOUND_MIXER_INFO, &info) != -1) {
	  label = g_strdup_printf(index ? "%s" : "Default (%s)", info.name);
	  add_menu_item(menu, label, signal_f, index);
	  g_free(label);
	  index++;
	}
	close(fd);
      }
      else break;
    }
#endif                                                                       
  }
  
  /* create default entry if no device(s) could be found */
  if(index == 0) add_menu_item(menu, "Default", signal_f, 0);

  /* attach menu */
  gtk_option_menu_set_menu(GTK_OPTION_MENU(option_menu), menu);
}

/*-- oss output callbacks ---------------------------------------------------*/

void config_adevice_cb(GtkWidget *widget, gint device)
{
  oss_audio_device = device;
}

void config_mdevice_cb(GtkWidget *widget, gint device)
{
  oss_mixer_device = device;
}

void on_config_adevice_alt_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  gboolean use_alt_audio_device = gtk_toggle_button_get_active(togglebutton);
  SET_SENSITIVE("oss_adevice_optionmenu", !use_alt_audio_device);
  SET_SENSITIVE("oss_adevice_alt_entry",   use_alt_audio_device);
}

void on_config_mdevice_alt_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  gboolean use_alt_mixer_device = gtk_toggle_button_get_active(togglebutton);
  SET_SENSITIVE("oss_mdevice_optionmenu", !use_alt_mixer_device);
  SET_SENSITIVE("oss_mdevice_alt_entry",   use_alt_mixer_device);
}

/*** plugin output ***********************************************************/

void
get_output_plugin_name()
{
  OutputPlugin *op = g_list_nth_data(get_output_list(), op_index);

  if(op_name) {
    g_free(op_name);
    op_name = NULL;
  }

  if((op == NULL) || (op->filename == NULL)) return;
  op_name = g_strdup(g_basename(op->filename));
}

static gchar *
strip(gchar *s)
{
  gchar *p;
  if(!s) return NULL;
  for(; *s == ' '; s++);
  if(!*s) return s;
  for(p = s+strlen(s)-1; *p == ' '; p--);
  *++p = 0;
  return s;
}

static void
update_plugin_options(gchar **config_string, gchar *name,
		      plugin_config_t *pc, gboolean save)
{
  plugin_config_t default_pc = DEFAULT_OP_CONFIG;

  gchar *buffer = NULL;
  gchar  out[1024];

  gboolean plugin_found = FALSE;
  gchar   *plugin, *next_plugin;
  gchar   *args;

  if(pc && !save) *pc = default_pc;
  if(!config_string || !*config_string || !name || !pc) {
    DEBUG(("[crossfade] update_plugin_options: missing arg!\n"));
    return;
  }

  buffer = g_strdup(*config_string);
  out[0] = 0;

  for(plugin = buffer; plugin; plugin = next_plugin) {
    if((next_plugin = strchr(plugin, ';'))) *next_plugin++ = 0;
    if((args = strchr(plugin, '='))) *args++ = 0;
    plugin = strip(plugin);
    if(!*plugin || !args || !*args) continue;

    if(save) {
      if(0 == strcmp(plugin, name)) continue;
      if(*out) strcat(out, "; ");
      strcat(out, plugin);
      strcat(out, "=");
      strcat(out, args);
      continue;
    }
    else if(strcmp(plugin, name)) continue;

    args = strip(args);
    sscanf(args, "%d,%d,%d",
	   &pc->throttle_enable,
	   &pc->max_write_enable,
	   &pc->max_write_len);
    pc->max_write_len &= -4;
    plugin_found = TRUE;
  }

  if(save) {
    /* only save if settings differ from defaults */
    if((  pc->throttle_enable  != default_pc.throttle_enable)
       ||(pc->max_write_enable != default_pc.max_write_enable)
       ||(pc->max_write_len    != default_pc.max_write_len)) {
      if(*out) strcat(out, "; ");
      sprintf(out + strlen(out), "%s=%d,%d,%d", name,
	      pc->throttle_enable ? 1 : 0,
	      pc->max_write_enable ? 1 : 0,
	      pc->max_write_len);
    }
    if(*config_string) g_free(*config_string);
    *config_string = g_strdup(out);
  }

  g_free(buffer);
}

static void
config_plugin_cb(GtkWidget *widget, gint index);

static gint
scan_plugins(GtkWidget *option_menu, gchar *selected)
{
  GtkWidget *menu = gtk_menu_new();
  GList     *list = g_list_first(get_output_list());
  gint      index = 0;
  gint  sel_index = -1;
  gint  def_index = -1;

  /* sanity check */
  if(selected == NULL) selected = "";

  /* parse module list */
  while(list) {
    OutputPlugin *op = (OutputPlugin *)list->data;
    GtkWidget  *item = gtk_menu_item_new_with_label(op->description);
    
    if(op == get_oplugin_info())  /* disable selecting ourselves */
      gtk_widget_set_sensitive(item, FALSE);
    else {
      if(def_index == -1) def_index = index;
      if(selected && !strcmp(g_basename(op->filename), selected))
	sel_index = index;
    }

    /* create menu item */
    gtk_signal_connect(GTK_OBJECT(item), "activate",
		       config_plugin_cb, (gpointer)index++);
    gtk_widget_show(item);
    gtk_menu_append(GTK_MENU(menu), item);

    /* advance to next module */
    list = g_list_next(list);
  }
    
  /* attach menu */
  gtk_option_menu_set_menu(GTK_OPTION_MENU(option_menu), menu);
  
  if(sel_index == -1) {
    DEBUG(("[crossfade] scan_plugins: plugin not found (\"%s\")\n", selected));
    return def_index;  /* use default (first entry) */
  }
  return sel_index;
}

/*-- plugin output callbacks ------------------------------------------------*/

static void
config_plugin_cb(GtkWidget *widget, gint index)
{
  OutputPlugin *op = g_list_nth_data(get_output_list(), index);

  /* get plugin options from gui */
  op_config.throttle_enable  = GET_TOGGLE("op_throttle_check");
  op_config.max_write_enable = GET_TOGGLE("op_maxblock_check");
  op_config.max_write_len    = GET_SPIN  ("op_maxblock_spin");

  /* update op_config_string (from op_config) */
  update_plugin_options(&op_config_string, op_name, &op_config, TRUE);

  /* select new plugin */
  op_index = index;
  get_output_plugin_name();

  /* update op_config (from op_config_string) */
  update_plugin_options(&op_config_string, op_name, &op_config, FALSE);

  /* update gui */
  SET_SENSITIVE("op_configure_button", op && (op->configure != NULL)); 
  SET_SENSITIVE("op_about_button",     op && (op->about != NULL));
  SET_TOGGLE   ("op_throttle_check",   op_config.throttle_enable);
  SET_TOGGLE   ("op_maxblock_check",   op_config.max_write_enable);
  SET_SPIN     ("op_maxblock_spin",    op_config.max_write_len); 
  SET_SENSITIVE("op_maxblock_spin",    op_config.max_write_enable);
}

void on_output_plugin_configure_button_clicked (GtkButton *button, gpointer user_data)
{
  GList      *list = get_output_list();
  OutputPlugin *op = g_list_nth_data(list, op_index);

  if((op == NULL) || (op->configure == NULL)) return;
  op->configure();
}

void on_output_plugin_about_button_clicked(GtkButton *button, gpointer user_data)
{
  GList      *list = get_output_list();
  OutputPlugin *op = g_list_nth_data(list, op_index);

  if((op == NULL) || (op->about == NULL)) return;
  op->about();
}

void on_op_maxblock_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  gboolean op_maxblock_check = gtk_toggle_button_get_active(togglebutton);
  SET_SENSITIVE("op_maxblock_spin", op_maxblock_check);
}

/*** crossfader **************************************************************/

static void xf_config_cb(GtkWidget *widget, gint index);
static void xf_type_cb  (GtkWidget *widget, gint index);

static void
create_crossfader_optionmenu()
{
  GtkWidget *optionmenu, *menu;

  if((optionmenu = lookup_widget(config_win, "xf_config_optionmenu"))) {
    menu = gtk_menu_new();
    add_menu_item(menu, "Automatic songchange", xf_config_cb, FADE_CONFIG_XFADE);
    add_menu_item(menu, "Manual songchange",    xf_config_cb, FADE_CONFIG_MANUAL);
    gtk_option_menu_set_menu(GTK_OPTION_MENU(optionmenu), menu);
  }

  if((optionmenu = lookup_widget(config_win, "xf_type_optionmenu"))) {
    menu = gtk_menu_new();
    add_menu_item(menu, "Reopen output device", xf_type_cb, FADE_TYPE_REOPEN);
    add_menu_item(menu, "Flush output device",  xf_type_cb, FADE_TYPE_FLUSH);
    add_menu_item(menu, "None (gapless)",       xf_type_cb, FADE_TYPE_NONE);
    add_menu_item(menu, "Pause",                xf_type_cb, FADE_TYPE_PAUSE);
    add_menu_item(menu, "Simple crossfade",     xf_type_cb, FADE_TYPE_SIMPLE_XF);
    add_menu_item(menu, "Advanced crossfade",   xf_type_cb, FADE_TYPE_ADVANCED_XF);
    gtk_option_menu_set_menu(GTK_OPTION_MENU(optionmenu), menu);
  }
}

static void
check_crossfader_dependencies()
{
  gboolean out_enable;
  gboolean in_locked;
  gboolean in_enable;

  /* HACK: avoid endless recursion */
  if(xf_checking_dependencies) return;
  xf_checking_dependencies = TRUE;
  
  out_enable = xf_config[xf_index].out_enable;
  in_locked  = xf_config[xf_index].in_locked;
  in_enable  = xf_config[xf_index].in_enable;

  SET_SENSITIVE("fadeout_length_spin",  out_enable);
  SET_SENSITIVE("fadeout_length_label", out_enable);
  SET_SENSITIVE("fadeout_volume_spin",  out_enable);
  SET_SENSITIVE("fadeout_volume_label", out_enable);

  SET_SENSITIVE("fadein_enable_check", !in_locked);
  SET_SENSITIVE("fadein_length_spin",  !in_locked && in_enable);
  SET_SENSITIVE("fadein_length_label", !in_locked && in_enable);
  SET_SENSITIVE("fadein_volume_spin",  !in_locked && in_enable);
  SET_SENSITIVE("fadein_volume_label", !in_locked && in_enable);

  if(in_locked) {
    SET_TOGGLE("fadein_enable_check", out_enable);
    SET_SPIN  ("fadein_length_spin",  xf_config[xf_index].out_len_ms);
    SET_SPIN  ("fadein_volume_spin",  xf_config[xf_index].out_volume);
  }
#if 1
  else {
    SET_TOGGLE("fadein_enable_check", in_enable);
    SET_SPIN  ("fadein_length_spin",  xf_config[xf_index].in_len_ms);
    SET_SPIN  ("fadein_volume_spin",  xf_config[xf_index].in_volume);
  }
#endif

  if(out_enable) SET_SENSITIVE("xfofs_lockout_radiobutton", TRUE);
  if(!in_locked && in_enable) SET_SENSITIVE("xfofs_lockin_radiobutton", TRUE);
  
  switch(xf_config[xf_index].ofs_type) {
  case FC_OFFSET_LOCK_OUT:
    if(!out_enable) SET_TOGGLE("xfofs_none_radiobutton", TRUE);
    break;
  case FC_OFFSET_LOCK_IN:
    if(!(!in_locked && in_enable))
      SET_TOGGLE((in_locked && out_enable)
		 ? "xfofs_lockout_radiobutton"
		 : "xfofs_none_radiobutton", TRUE);
    break;
  }

  switch(xf_config[xf_index].ofs_type_wanted) {
  case FC_OFFSET_LOCK_OUT:
    if(out_enable) SET_TOGGLE("xfofs_lockout_radiobutton", TRUE);
    break;
  case FC_OFFSET_LOCK_IN:
    if(!in_locked && in_enable) {
      SET_TOGGLE("xfofs_lockin_radiobutton", TRUE);
    }
    else if(out_enable) {
      SET_TOGGLE("xfofs_lockout_radiobutton", TRUE);
    }
    break;
  }

  if(!out_enable) SET_SENSITIVE("xfofs_lockout_radiobutton", FALSE);
  if(!(!in_locked && in_enable))
    SET_SENSITIVE("xfofs_lockin_radiobutton", FALSE);
  
  SET_SENSITIVE("xfofs_custom_spin",
		xf_config[xf_index].ofs_type == FC_OFFSET_CUSTOM);


  switch(xf_config[xf_index].ofs_type) {
  case FC_OFFSET_NONE:
    SET_SPIN("xfofs_custom_spin", 0);
    break;
  case FC_OFFSET_LOCK_OUT:
    SET_SPIN("xfofs_custom_spin", -xf_config[xf_index].out_len_ms);
    break;
  case FC_OFFSET_LOCK_IN:
    SET_SPIN("xfofs_custom_spin", -xf_config[xf_index].in_len_ms);
    break;
  case FC_OFFSET_CUSTOM:
    SET_SPIN("xfofs_custom_spin", xf_config[xf_index].ofs_custom_ms);
    break;
  }

  xf_checking_dependencies = FALSE;
}

/*-- crossfader callbacks ---------------------------------------------------*/

void
xf_config_cb(GtkWidget *widget, gint index)
{
  xf_index = index;

  SET_HISTORY("xf_type_optionmenu", xf_config[xf_index].type);
  SET_PAGE   ("xf_type_notebook",   xf_config[xf_index].type);

  SET_SPIN  ("pause_length_spin", xf_config[xf_index].pause_len_ms);

  SET_SPIN  ("simple_length_spin", xf_config[xf_index].simple_len_ms);

  SET_TOGGLE("fadeout_enable_check", xf_config[xf_index].out_enable);
  SET_SPIN  ("fadeout_length_spin",  xf_config[xf_index].out_len_ms);
  SET_SPIN  ("fadeout_volume_spin",  xf_config[xf_index].out_volume);
  
  SET_TOGGLE("fadein_lock_check",    xf_config[xf_index].in_locked);
  if(!xf_config[xf_index].in_locked) {
    SET_TOGGLE("fadein_enable_check", xf_config[xf_index].in_enable);
    SET_SPIN  ("fadein_length_spin",  xf_config[xf_index].in_len_ms);
    SET_SPIN  ("fadein_volume_spin",  xf_config[xf_index].in_volume);
  }

  switch(xf_config[xf_index].ofs_type) {
  case FC_OFFSET_NONE:
    SET_TOGGLE("xfofs_none_radiobutton", TRUE);
    break;
  case FC_OFFSET_LOCK_OUT:
    SET_TOGGLE("xfofs_lockout_radiobutton", TRUE);
    break;
  case FC_OFFSET_LOCK_IN:
    SET_TOGGLE("xfofs_lockin_radiobutton", TRUE);
    break;
  case FC_OFFSET_CUSTOM:
    SET_TOGGLE   ("xfofs_custom_radiobutton", TRUE);
    break;
  }      

  check_crossfader_dependencies();
}

void
xf_type_cb(GtkWidget *widget, gint index)
{
  SET_PAGE("xf_type_notebook", index);
  xf_config[xf_index].type = index;
}

void
on_pause_length_spin_changed(GtkEditable *editable, gpointer user_data)
{
  xf_config[xf_index].pause_len_ms
    = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
}

void
on_simple_length_spin_changed(GtkEditable *editable, gpointer user_data)
{
  xf_config[xf_index].simple_len_ms
    = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
}

void on_fadeout_enable_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  xf_config[xf_index].out_enable = gtk_toggle_button_get_active(togglebutton);
  check_crossfader_dependencies();
}

void on_fadeout_length_spin_changed(GtkEditable *editable, gpointer user_data)
{
  xf_config[xf_index].out_len_ms
    = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
  check_crossfader_dependencies();
}

void on_fadeout_volume_spin_changed(GtkEditable *editable, gpointer user_data)
{
  xf_config[xf_index].out_volume
    = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
  check_crossfader_dependencies();
}

void on_xfofs_none_radiobutton_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if(gtk_toggle_button_get_active(togglebutton)) {
    xf_config[xf_index].ofs_type = FC_OFFSET_NONE;
    if(!xf_checking_dependencies) {
      xf_config[xf_index].ofs_type_wanted = FC_OFFSET_NONE;
      check_crossfader_dependencies();
    }
  }
}

void on_xfofs_none_radiobutton_clicked(GtkButton *button, gpointer user_data)
{
  if(!xf_checking_dependencies)
    xf_config[xf_index].ofs_type_wanted = FC_OFFSET_NONE;
}

void on_xfofs_lockout_radiobutton_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if(gtk_toggle_button_get_active(togglebutton)) {
    xf_config[xf_index].ofs_type = FC_OFFSET_LOCK_OUT;
    if(!xf_checking_dependencies) {
      xf_config[xf_index].ofs_type_wanted = FC_OFFSET_LOCK_OUT;
      check_crossfader_dependencies();
    }
  }
}

void on_xfofs_lockout_radiobutton_clicked(GtkButton *button, gpointer user_data)
{
  if(!xf_checking_dependencies)
    xf_config[xf_index].ofs_type_wanted = FC_OFFSET_LOCK_OUT;
}

void on_xfofs_lockin_radiobutton_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if(gtk_toggle_button_get_active(togglebutton)) {
    xf_config[xf_index].ofs_type = FC_OFFSET_LOCK_IN;
    if(!xf_checking_dependencies) {
      xf_config[xf_index].ofs_type_wanted = FC_OFFSET_LOCK_IN;
      check_crossfader_dependencies();
    }
  }
}

void on_xfofs_custom_radiobutton_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if(gtk_toggle_button_get_active(togglebutton)) {
    xf_config[xf_index].ofs_type = FC_OFFSET_CUSTOM;
    if(!xf_checking_dependencies) {
      xf_config[xf_index].ofs_type_wanted = FC_OFFSET_CUSTOM;
      check_crossfader_dependencies();
    }
  }
}

void on_xfofs_custom_spin_changed(GtkEditable *editable, gpointer user_data)
{
  if(!xf_checking_dependencies)
    xf_config[xf_index].ofs_custom_ms = 
      gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
}

void on_fadein_lock_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  xf_config[xf_index].in_locked = gtk_toggle_button_get_active(togglebutton);
  check_crossfader_dependencies();
}

void on_fadein_enable_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if(!xf_checking_dependencies) {
    xf_config[xf_index].in_enable = gtk_toggle_button_get_active(togglebutton);
    check_crossfader_dependencies();
  }
}

void on_fadein_length_spin_changed(GtkEditable *editable, gpointer user_data)
{
  if(GTK_WIDGET_SENSITIVE(editable)) {
    xf_config[xf_index].in_len_ms
      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
    /* can't use check_crossfader_dependencies() here */
    if(xf_config[xf_index].ofs_type == FC_OFFSET_LOCK_IN)
      SET_SPIN("xfofs_custom_spin", -xf_config[xf_index].in_len_ms);
  }
}

void on_fadein_volume_spin_changed(GtkEditable *editable, gpointer user_data)
{
  if(GTK_WIDGET_SENSITIVE(editable))
    xf_config[xf_index].in_volume
      = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
}

/*** gap killer **************************************************************/

static void
check_gapkiller_dependencies()
{
  gboolean gap_lead_enable  = GET_TOGGLE("lgap_enable_check");
  gboolean gap_trail_enable = GET_TOGGLE("tgap_enable_check");
  gboolean gap_trail_locked = GET_TOGGLE("tgap_lock_check");

  if(gap_trail_locked) {
    gint gap_lead_length = GET_SPIN("lgap_length_spin");
    gint gap_lead_level  = GET_SPIN("lgap_level_spin");
    SET_TOGGLE("tgap_enable_check", gap_lead_enable);
    SET_SPIN  ("tgap_length_spin",  gap_lead_length);
    SET_SPIN  ("tgap_level_spin",   gap_lead_level);
  }

  SET_SENSITIVE("lgap_length_spin",  gap_lead_enable);
  SET_SENSITIVE("lgap_length_label", gap_lead_enable);
  SET_SENSITIVE("lgap_level_spin",   gap_lead_enable);
  SET_SENSITIVE("lgap_level_label",  gap_lead_enable);
  SET_SENSITIVE("tgap_enable_check", !gap_trail_locked);
  SET_SENSITIVE("tgap_length_spin",  !gap_trail_locked && gap_trail_enable);
  SET_SENSITIVE("tgap_length_label", !gap_trail_locked && gap_trail_enable);
  SET_SENSITIVE("tgap_level_spin",   !gap_trail_locked && gap_trail_enable);
  SET_SENSITIVE("tgap_level_label",  !gap_trail_locked && gap_trail_enable);
}

/*-- gapkiller callbacks ----------------------------------------------------*/

void on_lgap_enable_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  config->gap_lead_enable = gtk_toggle_button_get_active(togglebutton);
  SET_SENSITIVE("lgap_length_spin", config->gap_lead_enable);
  SET_SENSITIVE("lgap_level_spin",  config->gap_lead_enable);

  if(config->gap_trail_locked)
    SET_TOGGLE("tgap_enable_check", config->gap_lead_enable);
}

void on_tgap_lock_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  config->gap_trail_locked = GET_TOGGLE("tgap_lock_check");
  SET_SENSITIVE("tgap_enable_check", !config->gap_trail_locked);

  if(config->gap_trail_locked) {
    SET_SENSITIVE("tgap_length_spin",  FALSE);
    SET_SENSITIVE("tgap_level_spin",   FALSE);

    config->gap_trail_enable = GET_TOGGLE("tgap_enable_check");
    config->gap_trail_len_ms = GET_SPIN  ("tgap_length_spin");
    config->gap_trail_level  = GET_SPIN  ("tgap_level_spin");
    SET_TOGGLE   ("tgap_enable_check", config->gap_lead_enable);
    SET_SPIN     ("tgap_length_spin",  config->gap_lead_len_ms);
    SET_SPIN     ("tgap_level_spin",   config->gap_lead_level);
  }
  else {
    SET_TOGGLE   ("tgap_enable_check", config->gap_trail_enable);
    SET_SPIN     ("tgap_length_spin",  config->gap_trail_len_ms);
    SET_SPIN     ("tgap_level_spin",   config->gap_trail_level);

    SET_SENSITIVE("tgap_length_spin",  config->gap_trail_enable);
    SET_SENSITIVE("tgap_level_spin",   config->gap_trail_enable);
  }
}

void on_tgap_enable_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  if(GTK_WIDGET_SENSITIVE(togglebutton)) {
    config->gap_trail_enable = gtk_toggle_button_get_active(togglebutton);
    SET_SENSITIVE("tgap_length_spin", config->gap_trail_enable);
    SET_SENSITIVE("tgap_level_spin",  config->gap_trail_enable);
  }
}

void on_lgap_length_spin_changed(GtkEditable *editable, gpointer user_data)
{
  if(GTK_WIDGET_SENSITIVE(editable)) {
    config->gap_lead_len_ms = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
    if(config->gap_trail_locked)
      SET_SPIN("tgap_length_spin", config->gap_lead_len_ms);
  }
}

void on_lgap_level_spin_changed(GtkEditable *editable, gpointer user_data)
{
  if(GTK_WIDGET_SENSITIVE(editable)) {
    config->gap_lead_level = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(editable));
    if(config->gap_trail_locked)
      SET_SPIN("tgap_level_spin", config->gap_lead_level);
  }
}

/*** misc ********************************************************************/

/*-- misc callbacks ---------------------------------------------------------*/

void on_config_mixopt_enable_check_toggled(GtkToggleButton *togglebutton, gpointer user_data)
{
  gboolean enable_mixer_check = gtk_toggle_button_get_active(togglebutton);
  SET_SENSITIVE("mixopt_reverse_check",   enable_mixer_check);
}

/*** main config *************************************************************/

void on_config_apply_clicked(GtkButton *button, gpointer user_data)
{
  GtkWidget *widget;

  /* get current notebook page */
  if((widget = lookup_widget(config_win, "config_notebook")))
    config->notebook_page = gtk_notebook_get_current_page(GTK_NOTEBOOK(widget));

  /* output method */
  config->output_method = output_method;

  /* output method: builtin OSS */
  config->oss_audio_device = oss_audio_device;
  config->oss_use_alt_audio_device = GET_TOGGLE("oss_adevice_alt_check");
  if((widget = lookup_widget(config_win, "oss_adevice_alt_entry"))) {
    if(config->oss_alt_audio_device) g_free(config->oss_alt_audio_device);
    config->oss_alt_audio_device = gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1);
    g_strstrip(config->oss_alt_audio_device);
  }

  config->oss_mixer_device = oss_mixer_device;
  config->oss_use_alt_mixer_device = GET_TOGGLE("oss_mdevice_alt_check");
  if((widget = lookup_widget(config_win, "oss_mdevice_alt_entry"))) {
    if(config->oss_alt_mixer_device) g_free(config->oss_alt_mixer_device);
    config->oss_alt_mixer_device = gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1);
    g_strstrip(config->oss_alt_mixer_device);
  }
  config->oss_mixer_use_master = GET_TOGGLE("ossmixer_pcm_check");

  config->oss_buffer_size_ms = GET_SPIN("ossbuf_buffer_spin");
  config->oss_preload_size_ms = GET_SPIN("ossbuf_preload_spin");

  /* output method: plugin */
  op_config.throttle_enable  = GET_TOGGLE("op_throttle_check");
  op_config.max_write_enable = GET_TOGGLE("op_maxblock_check");
  op_config.max_write_len    = GET_SPIN  ("op_maxblock_spin");

  update_plugin_options(&op_config_string, op_name, &op_config, TRUE);

  if(config->op_name) g_free(config->op_name);
  config->op_name = op_name ? strdup(op_name) : NULL;

  if(config->op_config_string) g_free(config->op_config_string);
  config->op_config_string = op_config_string ? strdup(op_config_string) : NULL;

  config->op_config = op_config;

  /* output method: none: */

  /* crossfader */
  config->mix_size_ms = GET_SPIN("xf_buffer_spin");

  config->crossfade_config_index = xf_index;
  memcpy(config->fc, xf_config, sizeof(xf_config));
  
  /* gap killer */
  config->gap_lead_enable  = GET_TOGGLE("lgap_enable_check");
  config->gap_lead_len_ms  = GET_SPIN  ("lgap_length_spin");
  config->gap_lead_level   = GET_SPIN  ("lgap_level_spin");
  config->gap_trail_locked = GET_TOGGLE("tgap_lock_check");
  if(config->gap_trail_locked) {
    config->gap_trail_enable = unlocked_gap_trail_enable;
    config->gap_trail_len_ms = unlocked_gap_trail_len_ms;
    config->gap_trail_level  = unlocked_gap_trail_level;
  }
  else {
    config->gap_trail_enable = GET_TOGGLE("tgap_enable_check");
    config->gap_trail_len_ms = GET_SPIN  ("tgap_length_spin");
    config->gap_trail_level  = GET_SPIN  ("tgap_level_spin");
  }

  /* misc */
  config->enable_debug   = GET_TOGGLE("debug_stderr_check");
  config->enable_monitor = GET_TOGGLE("debug_monitor_check");
  config->enable_mixer   = GET_TOGGLE("mixopt_enable_check");
  config->mixer_reverse  = GET_TOGGLE("mixopt_reverse_check"); 

  /* save configuration */
  xfade_save_config();
  
  /* show/hide monitor win depending on config->enable_monitor */
  xfade_check_monitor_win(); 
}

void on_config_ok_clicked(GtkButton *button, gpointer user_data)
{
  /* apply and save config */
  on_config_apply_clicked(button, user_data);

  /* close and destroy window */
  gtk_widget_destroy(config_win);
}
void xfade_configure()
{
  GtkWidget *widget;

  if(!config_win) {
    /* create */
    if(!(config_win = create_config_win())) {
      DEBUG(("[crossfade] plugin_configure: error creating window!\n"));
      return;
    }

    /* update config_win when window is destroyed */
    gtk_signal_connect(GTK_OBJECT(config_win), "destroy",
		       GTK_SIGNAL_FUNC(gtk_widget_destroyed), &config_win);
    
    /* go to remembered notebook page */
    if((widget = lookup_widget(config_win, "config_notebook")))
      gtk_notebook_set_page(GTK_NOTEBOOK(widget), config->notebook_page);

    /* output method */
    output_method = config->output_method;
    switch(output_method) {
    case OUTPUT_METHOD_BUILTIN_OSS:
      widget = lookup_widget(config_win, "output_oss_radio");
      break;
    case OUTPUT_METHOD_PLUGIN:
      widget = lookup_widget(config_win, "output_plugin_radio");
      break;
    case OUTPUT_METHOD_BUILTIN_NULL:
      widget = lookup_widget(config_win, "output_none_radio");
      break;
    default:
      widget = NULL;
    }
    if(widget) gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget), TRUE);

    if((widget = lookup_widget(config_win, "output_notebook")))
      gtk_notebook_set_page(GTK_NOTEBOOK(widget), output_method);

    /* output method: builtin OSS */
    oss_audio_device = config->oss_audio_device;
    oss_mixer_device = config->oss_mixer_device;

    if((widget = lookup_widget(config_win, "oss_adevice_optionmenu"))) {
      scan_devices("Audio devices:", widget, config_adevice_cb);
      gtk_option_menu_set_history(GTK_OPTION_MENU(widget), oss_audio_device);
      gtk_widget_set_sensitive(widget, !config->oss_use_alt_audio_device);
    }
    SET_TOGGLE("oss_adevice_alt_check", config->oss_use_alt_audio_device);
    if((widget = lookup_widget(config_win, "oss_adevice_alt_entry"))) {
      gtk_entry_set_text(GTK_ENTRY(widget), config->oss_alt_audio_device
			 ? config->oss_alt_audio_device
			 : DEFAULT_OSS_ALT_AUDIO_DEVICE);
      gtk_widget_set_sensitive(widget, config->oss_use_alt_audio_device);
    }

    if((widget = lookup_widget(config_win, "oss_mdevice_optionmenu"))) {
      scan_devices("Mixers:", widget, config_mdevice_cb);
      gtk_option_menu_set_history(GTK_OPTION_MENU(widget), oss_mixer_device);
      gtk_widget_set_sensitive(widget, !config->oss_use_alt_mixer_device);
    }
    SET_TOGGLE("oss_mdevice_alt_check", config->oss_use_alt_mixer_device);
    if((widget = lookup_widget(config_win, "oss_mdevice_alt_entry"))) {
      gtk_entry_set_text(GTK_ENTRY(widget), config->oss_alt_mixer_device
        		 ? config->oss_alt_mixer_device
        		 : DEFAULT_OSS_ALT_MIXER_DEVICE); 
      gtk_widget_set_sensitive(widget, config->oss_use_alt_mixer_device);
    }
    SET_TOGGLE("ossmixer_pcm_check", config->oss_mixer_use_master);

    SET_SPIN("ossbuf_buffer_spin",  config->oss_buffer_size_ms);
    SET_SPIN("ossbuf_preload_spin", config->oss_preload_size_ms);

    /* output method: plugin */
    if(op_name) g_free(op_name);
    op_name = config->op_name ? strdup(config->op_name) : NULL;

    if(op_config_string) g_free(op_config_string);
    op_config_string = config->op_config_string
      ? strdup(config->op_config_string) : NULL;

    update_plugin_options(&op_config_string, op_name, &op_config, FALSE);

    SET_TOGGLE   ("op_throttle_check",   op_config.throttle_enable);
    SET_TOGGLE   ("op_maxblock_check",   op_config.max_write_enable);
    SET_SPIN     ("op_maxblock_spin",    op_config.max_write_len); 
    SET_SENSITIVE("op_maxblock_spin",    op_config.max_write_enable);

    if((widget = lookup_widget(config_win, "op_plugin_optionmenu"))) {
      OutputPlugin *op = NULL;
      if((op_index = scan_plugins(widget, config->op_name)) >= 0) {
	gtk_option_menu_set_history(GTK_OPTION_MENU(widget), op_index);
	op = g_list_nth_data(get_output_list(), op_index);
      }
      SET_SENSITIVE("op_configure_button", op && (op->configure != NULL));
      SET_SENSITIVE("op_about_button",     op && (op->about     != NULL));
    }
    
    /* crossfader */
    create_crossfader_optionmenu();

    SET_SPIN("xf_buffer_spin", config->mix_size_ms);

    xf_index = CLAMP(config->crossfade_config_index, 0, MAX_FADE_CONFIGS);
    memcpy(xf_config, config->fc, sizeof(xf_config));
    SET_HISTORY("xf_config_optionmenu", xf_index);
    xf_config_cb(NULL, xf_index);

    check_crossfader_dependencies();
    
    /* gap killer */
    SET_TOGGLE   ("lgap_enable_check", config->gap_lead_enable);
    SET_SPIN     ("lgap_length_spin",  config->gap_lead_len_ms);
    SET_SPIN     ("lgap_level_spin",   config->gap_lead_level);
    SET_TOGGLE   ("tgap_lock_check",   config->gap_trail_locked);
    SET_TOGGLE   ("tgap_enable_check", config->gap_trail_enable);
    SET_SPIN     ("tgap_length_spin",  config->gap_trail_len_ms);
    SET_SPIN     ("tgap_level_spin",   config->gap_trail_level);

    unlocked_gap_trail_enable = config->gap_trail_enable;
    unlocked_gap_trail_len_ms = config->gap_trail_len_ms;
    unlocked_gap_trail_level  = config->gap_trail_level;

    check_gapkiller_dependencies();

    /* misc */
    SET_TOGGLE("debug_stderr_check", config->enable_debug);
    SET_TOGGLE("debug_monitor_check", config->enable_monitor);
    SET_TOGGLE("mixopt_enable_check", config->enable_mixer);
    SET_TOGGLE("mixopt_reverse_check", config->mixer_reverse);

    /* show window near mouse pointer */
    gtk_window_set_position(GTK_WINDOW(config_win), GTK_WIN_POS_MOUSE);
    gtk_widget_show(config_win);
  }
  else
    /* bring window to front */
    gdk_window_raise(config_win->window);
}

void xfade_about()
{
  if(!about_win) {
    about_win = create_about_win();

    /* update about_win when window is destroyed */
    gtk_signal_connect(GTK_OBJECT(about_win), "destroy", GTK_SIGNAL_FUNC(gtk_widget_destroyed), &about_win);
    
    /* show near mouse pointer */
    gtk_window_set_position(GTK_WINDOW(about_win), GTK_WIN_POS_MOUSE);
    gtk_widget_show(about_win);
  }
  else
    gdk_window_raise(about_win->window);
}
